/************************/
/* Most common routines */
/************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stack-c.h"
#include "common.h"

/* 
 * Print some info about image 
 */
void printinfo(char *filename, BitmapFileHeader *bfh, BitmapInfoHeader *bih) {
    sciprint("----------------------------------------\n");
    sciprint("Bitmap info about %s:\n", filename);
    sciprint("----------------------------------------\n");
    sciprint("file type:        %u\n", bfh->bfType);
    sciprint("file size:        %u\n", bfh->bfSize);
    sciprint("bfReserved1:      %u\n", bfh->bfReserved1);
    sciprint("bfReserved2:      %u\n", bfh->bfReserved2);
    sciprint("data offset:      %u\n", bfh->bfOffBits);
    sciprint("----------------------------------------\n");
    sciprint("bmp info size:    %u\n", bih->biSize);
    sciprint("bmp width:        %u\n", bih->biWidth);
    sciprint("bmp height:       %u\n", bih->biHeight);
    sciprint("bmp planes:       %u\n", bih->biPlanes);
    sciprint("bmp bit count:    %u\n", bih->biBitCount);
    sciprint("commpression:     %u\n", bih->biCompression);
    sciprint("bmp image size:   %u\n", bih->biSizeImage);
    sciprint("XPelsPerMeter:    %u\n", bih->biXPelsPerMeter);
    sciprint("YPelsPerMeter:    %u\n", bih->biYPelsPerMeter);
    sciprint("bmp color used:   %u\n", bih->biClrUsed);
    sciprint("color important:  %u\n", bih->biClrImportant);
    sciprint("----------------------------------------\n");
}

/* 
 * Creates new hypermatrix structure to be further passed to scilab 
 * using cre_hmat
 */
HyperMat *new_sci_tru_img(unsigned rows, unsigned cols) {
    HyperMat *h;

    h = (HyperMat *) malloc(sizeof(HyperMat));
    if (!h) error("unable to alloc memory for the return image\n");

    h->size = 3*rows*cols;
    h->it = 0; // real

    h->R = (double *)malloc(h->size*sizeof(double));
    if (!h->R) error("unable to alloc memory for the return image\n");
    h->I = NULL;

    h->sc.m = 1;
    h->sc.n = 3;
    h->sc.l = -1; // dimensions matrix not in scilab stack
    h->sc.it = I_INT32;
    h->sc.D = malloc(3*sizeof(I_INT32));
    if (!h->sc.D) error("unable to alloc memory for the return image\n");

    IC_INT32(h->sc.D)[0] = rows;
    IC_INT32(h->sc.D)[1] = cols;
    IC_INT32(h->sc.D)[2] = 3;

    return h;
}

/*
 * Creates a Scilab hypermatrix (mlist)
 * from a hypermatrix structure 
 */
int cre_hmat(int pos, HyperMat *H) {
    static char *Str[]= { "hm","dims","entries"}; int m1=1,n1=3;
    int mL=3,nL=1,lL, un=1;
  
    CreateVar(pos,"m", &mL, &nL, &lL);
    CreateListVarFromPtr(pos,1,"S", &m1, &n1, Str);
    CreateListVarFromPtr(pos,2,"I",&(H->sc.m), &(H->sc.n), &(H->sc));
    CreateListCVarFromPtr(pos,3,"d", &H->it,&H->size,&un ,&H->R,&H->I);
    return 1;
}

/*
 * Creates a hypermatrix structure from a Scilab
 * hypermatrix (mlist)   
 */
int get_rhs_hmat(int pos, HyperMat *H) {
    char **Str2;
    int m,n,l,m1,n1,m2,n2,m3,n3,l3,it3,lc3;
    GetRhsVar(pos,"m",&m,&n,&l);
    CheckLength(pos,m,3);
    GetListRhsVar(pos,1,"S",&m1,&n1,&Str2);
    if ( strcmp(Str2[0],"hm") != 0) {
	Scierror(999,"Argument %d is not an hypermatrix\r\n",pos);
	return 0;
    }
    GetListRhsVar(pos,2,"I",&m2,&n2,&H->sc);
    GetListRhsCVar(pos,3,"d",&it3,&m3,&n3,&l3,&lc3);
    H->size = m3*n3;
    H->R = stk(l3);
    H->I = (it3==1) ? stk(lc3): NULL;
    H->it= it3;
    FreeRhsSVar(Str2);
    return 1;
}

/*
 * Get true image from input params
 */
int get_rhs_tru_img(int pos, HyperMat *h, char *filename) {
    GetRhsHMat(pos,h);
    if (h->sc.n != 3) {error("argument 1 is not a 3D hypermatrix");}
	else if (IC_INT32(h->sc.D)[2] != 3) {error("argument 1 is not a M x N x 3 hypermatrix");}
    return 1;
}

/*
 * Free garbage
 */
void free_sci_tru_img(HyperMat *h) {

    free(h->sc.D);
    free(h->R);
    free (h);
    h = NULL;
}

/*
 * Read 1-bit bitmap
 */
int imread1(byte *buf) {
    int *out;			// Output variable
    int xout, yout;		// Output dimensions
    dword x, y;			// Coordinates of pixel
    dword count, lim;		// Count and limit
    size_t pos = 0;		// Current position in buffer
    byte mask;			// Mask
    dword k=0;

    /* Allocate memory for output array */
    if ((out = (int *) malloc((unsigned) bih.biWidth*bih.biHeight * sizeof(int))) == NULL) {
	error("Can't allocate memory for output data\n");
        if(buf != NULL)
            free(buf);
	return 1;
    }

    /* Determine dimensions */
    xout = bih.biHeight;
    yout = bih.biWidth;
    
    /* Starting coordinates */
    x = bih.biHeight - 1;
    y = 0;

    /* y-dimension of output array */
    lim = bih.biWidth/8*8 + bih.biWidth%8;

    /* Main part of programm - reading pixels */
    while (pos < bih.biSizeImage) {
        count = 7;
        mask = 0x80;
    	while (y < lim) {
	    OUT(x,y++) = (buf[pos] & mask) >> count;
	    count -= 1;
            mask >>= 1;
            if (count == -1) {
                k++;
                pos++;
                count = 7;
                mask = 0x80;
    	    }
	}
	pos += 4 - k%4;
        x--;
        y = 0;
        k = 0;
    }

    /* Create an out array */
    CreateVarFromPtr(2, "i", &xout, &yout, &out);
    FreePtr(&out);

    return 0;
}

/*
 * Read 4-bit bitmap
 */
int imread4(byte *buf) {
    int *out;			// Output variable
    int xout, yout;		// Output dimensions
    dword x, y;			// Coordinates of pixel
    dword count, lim, shift;	// Count, limit and shift
    size_t pos = 0;		// Current position in buffer
    byte mask;			// Mask

    /* Determine dimensions */
    xout = bih.biHeight;
    yout = bih.biWidth;
    
    /* Allocate memory for output array */
    if ((out = (int *) malloc((unsigned) xout*yout * sizeof(int))) == NULL) {
	error("Can't allocate memory for output data\n");
        if(buf != NULL)
            free(buf);
	return 1;
    }
	
    /* Starting coordinates */
    x = bih.biHeight - 1;
    y = 0;

    /* y-dimension of output array */
    lim = yout/8*8 + yout%8;

    /* Determine shift */
    shift = (lim/2%4==0)?0:(4-lim/2%4);

    /* Main part of programm - reading pixels */
    while (pos < bih.biSizeImage) {
        count = 4;
        mask = 0xF0;
    	while (y < lim) {
	    OUT(x,y++) = (buf[pos] & mask) >> count;
	    count -= 4;
            mask >>= 4;
            if (count == -4) {
                pos++;
                count = 4;
                mask = 0xF0;
    	    }
	}
	pos += shift;
        x--;
        y = 0;
    }

    if (Lhs == 2) {
	
    }
    
    /* Create an out array */
    CreateVarFromPtr(2, "i", &xout, &yout, &out);
    FreePtr(&out);

    return 0;
}

/*
 * Read 8-bit bitmap
 */
int imread8(byte *buf) {
    int *out;			// Output variable
    int xout, yout;		// Output dimensions
    dword x, y;			// Coordinates of pixel
    dword lim, shift;		// Limit and shift
    size_t pos = 0;		// Current position in buffer

    /* Determine dimensions */
    xout = bih.biHeight;
    yout = bih.biWidth;
    
    /* Allocate memory for output array */
    if ((out = (int *) malloc((unsigned) xout*yout * sizeof(int))) == NULL) {
	error("Can't allocate memory for output data\n");
        if(buf != NULL)
            free(buf);
	return 1;
    }

    /* Starting coordinates */
    x = bih.biHeight - 1;
    y = 0;

    /* y-dimension of output array */
    lim = yout/8*8 + yout%8;
    
    shift = (lim%4==0)?0:(4-lim%4);

    /* Main part of programm - reading pixels */
    while (pos < bih.biSizeImage) {
    	while (y < lim) {
	    OUT(x,y++) = buf[pos];
            pos++;
    	}
	pos += shift;
        x--;
        y = 0;
    }

    /* Create an out array */
    CreateVarFromPtr(2, "i", &xout, &yout, &out);
    FreePtr(&out);
    
    return 0;
}

/*
 * Read 24-bit bitmap
 */
int imread24(byte *buf) {
    HyperMat *H;		// Hyper matrix
    dword x, y;			// Coordinates of pixel
    size_t pos = 0;		// Current position in buffer

    /* Starting coordinates */
    x = bih.biHeight - 1;
    y = 0;

    /* Create new hyper matrix */
    H = (HyperMat *) new_sci_tru_img(bih.biHeight,bih.biWidth);
    if (!H) {
	error("unable to alloc memory for the return image\n");
	return 1;
    }

    /* Main part of programm - reading pixels */
    while (pos < bih.biSizeImage) {
        while (y < bih.biWidth) {
            OUT3d(x,y,0) = buf[pos++];
            OUT3d(x,y,1) = buf[pos++];
            OUT3d(x,y,2) = buf[pos++];
	    if ((y+1)%bih.biWidth==0) pos += bih.biWidth%4;
            y++;
	}
    	x--;
	y = 0;
    }
    
    /* Create new hyper matrix variable */
    CreateHMat(2,H);
    
    /* Free garbage */
    free_sci_tru_img(H);
    
    return 0;
}

/*
 * Write 1-bit image
 */
int imwrite1(byte *buf, int *in) {
    dword x,y;			// Coordinates of pixel
    dword count, lim;		// Count and limit
    size_t pos = 0;		// Current position in buffer
    dword k=0;

    /* y-dimension of output array */
    lim = bih.biWidth/8*8 + bih.biWidth%8;

    /* Starting coordinates */
    x = bih.biHeight-1;
    y = 0;

    /* Main part of programm - writing bitmap data in buffer */
    memset(buf, 0, bih.biSizeImage);
    while (pos < bih.biSizeImage) {
	count = 7;
        while (y < lim) {
	    buf[pos] |= IN(x,y++) << count;
	    count -= 1;
            if (count == -1) {
    		k++;
        	count = 7;
        	pos++;
    	    }
    	}
    	pos += 4 - k%4;
	x--;
	y = 0;
	k = 0;
    }
    
    return 0;
}

/*
 * Write 4-bit image
 */
int imwrite4(byte *buf, int *in) {
    dword x,y;			// Coordinates of pixel
    dword count, lim, shift;	// Count, limit and shift
    size_t pos = 0;		// Current position in buffer
    
    /* y-dimension of output array */
    lim = bih.biWidth/8*8 + bih.biWidth%8;

    /* Determine shift */
    shift = (lim/2%4==0)?0:(4-lim/2%4);

    /* Starting coordinates */
    x = bih.biHeight-1;
    y = 0;

    /* Main part of programm - writing bitmap data in buffer */
    memset(buf, 0, bih.biSizeImage);
    while (pos < bih.biSizeImage) {
	count = 4;
        while (y < lim) {
    	    buf[pos] |= IN(x,y++) << count;
	    count -= 4;
            if (count == -4) {
        	count = 4;
        	pos++;
            }
    	}
	pos += shift;
    	x--;
    	y = 0;
    }
    
    return 0;
}

/*
 * Write 8-bit image
 */
int imwrite8(byte *buf, int *in) {
    dword x,y;			// Coordinates of pixel
    dword count, lim, shift;	// Count, limit and shift
    size_t pos = 0;		// Current position in buffer
    
    /* y-dimension of output array */
    lim = bih.biWidth/8*8 + bih.biWidth%8;

    /* Determine shift */
    shift = (lim%4==0)?0:(4-lim%4);

    /* Starting coordinates */
    x = bih.biHeight-1;
    y = 0;

    /* Main part of programm - writing bitmap data in buffer */
    memset(buf, 0, bih.biSizeImage);
    while (pos < bih.biSizeImage) {
	count = 0;
        while (y < lim) {
	    buf[pos] = IN(x,y++);
	    pos++;
        }
	pos += shift;
    	x--;
    	y = 0;
    }

    return 0;
}

/*
 * Write 24-bit image
 */
int imwrite24(byte *buf, HyperMat *H) {
    dword x,y;			// Coordinates of pixel
    size_t pos = 0;		// Current position in buffer
    
    /* Starting coordinates */
    x = bih.biHeight-1;
    y = 0;

    /* Main part of programm - writing bitmap data in buffer */
    memset(buf, 0, bih.biSizeImage);
    while (pos < bih.biSizeImage) {
	while (y < bih.biWidth) {
    	    buf[pos++] = IN3d(x,y,0);
            buf[pos++] = IN3d(x,y,1);
            buf[pos++] = IN3d(x,y,2);
	    if ((y+1)%bih.biWidth==0) pos += bih.biWidth%4;
    	    y++;
	}
	x--;
    	y = 0;
    }
    
    return 0;
}

/*
 * Read color map
 */
int index_map_to_sci(RGBquad *colormap, int nv) {
   double *map;
   unsigned i;
   double factor = 1.0 / MaxRGB;
   int rmap, cmap;

   rmap = bih.biClrUsed; cmap = 3;
   map = (double *)calloc(rmap*3, sizeof(double));
   if (!map) error("unable to alloc memory for output colormap");

   for (i=0; i < (unsigned) rmap; i++) {
      map[i]        = colormap[i].rgbRed   * factor;
      map[i+rmap]   = colormap[i].rgbGreen * factor;
      map[i+2*rmap] = colormap[i].rgbBlue  * factor;
   }

   CreateVarFromPtr(nv, "d",&rmap,&cmap,&map);  // colormap
   
   free(map);
   return 1;
}

/*
 * Write color map
 */
int sci_index_map_to_colormap(int nv, RGBquad *colormap) {
   unsigned i;
   int rmap, cmap, pmap, bits;

   GetRhsVar(nv, "d", &rmap, &cmap, &pmap);
   if(cmap != 3) error("colormap array must have exactly 3 columns");

   bih.biClrUsed=rmap;
   bih.biClrImportant=rmap;
   
   if (bih.biClrUsed > 15) bits = 8;
    else bits = 4;

   for (i=0; i<bih.biClrUsed; i++) {
      colormap[i].rgbRed = stk(pmap)[i] * MaxRGB;
      colormap[i].rgbGreen = stk(pmap)[i+bih.biClrUsed] * MaxRGB;
      colormap[i].rgbBlue  = stk(pmap)[i+2*(bih.biClrUsed)] * MaxRGB;
   } 

   return bits;
}
