/****************/
/* Common units */
/****************/

/* Bitmap signature 'BM' */
#define BMAGIC 0x4d42

/* Maximum value for color's components */
#define MaxRGB 255

/* Output and input arrays */
#define OUT(i,j) out[(i) + (j)*(xout)]
#define IN(i,j) in[(i) + (j)*(bih.biHeight)]
#define OUT3d(i,j,k) H->R[(i) + (j)*(bih.biHeight) + (k)*(bih.biHeight)*(bih.biWidth)]
#define IN3d(i,j,k) H->R[(i) + (j)*(bih.biHeight) + (k)*(bih.biHeight)*(bih.biWidth)]

/* Error messages */
#define error(str) {Scierror(999,"%s.\n\r", (str)); fprintf(stderr,"%s.\n", (str));}
#define warning(str) sciprint("%s (warning): %s.\n\r",filename, (str));

/* Some work with hypermatrices */
#define CreateHMat(pos,H) if (!cre_hmat(pos,H)) { return 0;}
#define GetRhsHMat(pos,H) if (!get_rhs_hmat(pos,H)) { return 0;}

/* Base types */
typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;

#pragma pack(push, 1)

/* File Header structure */
typedef struct {
    word	bfType;
    dword	bfSize;
    word	bfReserved1;
    word	bfReserved2;
    dword	bfOffBits;
} BitmapFileHeader;

/* Info Header structure */
typedef struct {
    dword	biSize;
    dword	biWidth;
    dword	biHeight;
    word	biPlanes;
    word	biBitCount;
    dword	biCompression;
    dword	biSizeImage;
    dword	biXPelsPerMeter;
    dword	biYPelsPerMeter;
    dword	biClrUsed;
    dword	biClrImportant;
} BitmapInfoHeader;

/* RGBquad structure */
typedef struct {
    byte	rgbBlue;
    byte	rgbGreen;
    byte	rgbRed;
    byte	rgbReserved;
} RGBquad;

/* Hyper matrix */
typedef struct {
  SciIntMat sc;   /* coding informations */
  int    it,      /* it=1 for complex; it=0 for real */
         size;    /* size of data array              */
  double *R,      /* real part                       */
         *I;      /* imaginary part (NULL for real hypermatrices) */
} HyperMat;

#pragma pack(pop)

/* Common variables */
BitmapFileHeader bfh;
BitmapInfoHeader bih;

/* Input and output files */
FILE *in;
FILE *out;

/* Functions from common.c */

/* Print some info about image */
void printinfo(char *filename, BitmapFileHeader *bfh, BitmapInfoHeader *bih);

/* Read 1-bit bmp */
int imread1(byte *buf);
/* Read 4-bit bmp */
int imread4(byte *buf);
/* Read 8-bit bmp */
int imread8(byte *buf);
/* Read 24-bit bmp */
int imread24(byte *buf);

/* Write 1-bit bmp */
int imwrite1(byte *buf, int *in);
/* Write 4-bit bmp */
int imwrite4(byte *buf, int *in);
/* Write 8-bit bmp */
int imwrite8(byte *buf, int *in);
/* Write 24-bit bmp */
int imwrite24(byte *buf, HyperMat *H);

/* New hypermatrix for true image */
HyperMat *new_sci_tru_img(unsigned rows, unsigned cols);
/* Create hypermatrix */
int cre_hmat(int pos, HyperMat *H);
/* Get true image */
int get_rhs_tru_img(int pos, HyperMat *h, char *filename);
/* Get hypermatrix*/
int get_rhs_hmat(int pos, HyperMat *H);
/* Free garbage */
void free_sci_tru_img(HyperMat *h);
/* Load colormap to Scilab */
int index_map_to_sci(RGBquad *colormap, int nv);
/* Get colormap from Scilab */
int sci_index_map_to_colormap(int nv, RGBquad *colormap);
