function [f] = filter2(h,image)
//
// Two-dimensional digital filtering.
// 
// Y = filter2(h,X)  filters the data in X with the 
// two-dimensional FIR filter in the matrix h. It computes
// the result, Y, using two-dimensional correlation, and 
// returns the central part of the correlation that is the
// same size as X.
//
// Created by D. Ushakov (dushakov@gmail.com)
//

// Check the mask
if type(h) ~= 1 then
  error('First argument must be a matrix');
end;
sh = size(h);
if sh(1) ~= sh(2) then
  error('Error dimensions of mask');
end;
dsh = fix(sh(1)/2);
if dsh == sh(1)/2 then
  error('Mask must be an even');
end;

// Window
sh = size(h);
win = zeros(sh(1),sh(2));
f = image;
s = size(image);
mx = image;
printf('This may take a while...');
select type(image)
  case 1 then   // 2D array
    for i = dsh+1:s(1)-dsh
      for j = dsh+1:s(2)-dsh
        // Move current window
        w1 = 1;
        for di = -dsh:dsh
          w2 = 1;
          for dj = -dsh:dsh
            win(w1,w2) = image(i+di,j+dj)*h(w1,w2);
            w2 = w2+1;
          end;
          w1 = w1+1;
        end;
        mx(i,j) = uint8(sum(win));
      end;
    end;
///////////////////////////////    
    h=-h;
    win = zeros(sh(1),sh(2));
    my = image;
    for i = dsh+1:s(1)-dsh
      for j = dsh+1:s(2)-dsh
        // Move current window
        w1 = 1;
        for di = -dsh:dsh
          w2 = 1;
          for dj = -dsh:dsh
            win(w1,w2) = image(i+di,j+dj)*h(w1,w2);
            w2 = w2+1;
          end;
          w1 = w1+1;
        end;
        my(i,j) = uint8(sum(win));
      end;
    end;  
    f=sqrt(mx.*mx + my.*my);
///////////////////////////////    
  case 17 then    // 3D array
    if s(3) ~= 3 then
      error('RGB image must have 3rd dimension equal to 3.');
    end;
    for k = 1:3
      for i = dsh+1:s(1)-dsh
        for j = dsh+1:s(2)-dsh
          // Move current window
          w1 = 1;
          for di = -dsh:dsh
            w2 = 1;
            for dj = -dsh:dsh
              win(w1,w2) = image(i+di,j+dj,k)*h(w1,w2);
              w2 = w2+1;
            end;
            w1 = w1+1;
          end;
          f(i,j,k) = uint8(sum(win));
        end;
      end;
    end;
  else
    error('Argument of invalid type.');
end;
endfunction
