/*************************/
/* Read image in array   */
/* a=imread("filename"); */
/*************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stack-c.h"
#include "common.h"

int imread(char *fname) {
    static int filename, m1, n1;
    static int minlhs=1, maxlhs=2, minrhs=1, maxrhs=1;
    
    /* Buffer */
    byte *buf = NULL;

    /* Color map */
    RGBquad *colormap;
    
    /* Check number of inputs and outputs */
    CheckRhs(minrhs,maxrhs);
    CheckLhs(minlhs,maxlhs);

    /* Get filename */
    GetRhsVar(1, "c", &m1, &n1, &filename);
    
    /* Open file */
    if ((in = fopen(cstk(filename), "rb")) == NULL) {
	error("Error opening input bmp file");
	return 0;
    }

    /* Read file header */
    if (fread(&bfh, sizeof(BitmapFileHeader), 1, in) != 1) {
	error("Error reading bmp file header");
        fclose(in);
        return 0;
    }
    
    /* Check filetype */
    if (bfh.bfType != BMAGIC) {
        error("Not a bitmap file");
        fclose(in);
        return 0;
    }
    
    /* Read bitmap info header */
    if (fread(&bih, sizeof(BitmapInfoHeader), 1, in) != 1) {
	error("Error reading bmp header");
        fclose(in);
        return 0;
    }

    /* Print some info about image */
    printinfo(cstk(filename), &bfh, &bih);

    if (Lhs == 2) {
	/* Allocate memory for colormap */
	if ((colormap = (RGBquad *) malloc(bih.biClrUsed*sizeof(RGBquad))) == NULL) {
    	    error("Can't allocate memory for colormap");
	    fclose(in);
	    return 0;
	}
	/* Read bitmap info header */
	if (fread(colormap, bih.biClrUsed*sizeof(RGBquad), 1, in) != 1) {
	    error("Error reading colormap");
    	    fclose(in);
    	    return 0;
	}
    }
    
    /* Look up for data */
    if (fseek(in, bfh.bfOffBits, SEEK_SET)) {
        error("Error openning data structure");
        fclose(in);
	return 0;
    }

    /* Allocate memory for buffer */
    if ((buf = (byte *) malloc(bih.biSizeImage)) == NULL) {
        error("Can't allocate memory for buffer");
	fclose(in);
	return 0;
    }

    /* Read bitmap data */
    memset(buf, 0, bih.biSizeImage);
    if (fread(buf, bih.biSizeImage, 1, in) != 1) {
        if(buf != NULL)
            free(buf);
	fclose(in);
	return 0;
    }

    /* Close input file */
    fclose(in);

    switch (bih.biBitCount) {
	case 1:
	    if (imread1(buf) != 0) {
		error("Error read data!!!");
		return 0;
	    }
	    break;
        case 4:
	    if (imread4(buf) != 0) {
		error("Error read data!!!");
		return 0;
	    }
	    break;
        case 8:
	    if (imread8(buf) != 0) {
		error("Error read data!!!");
		return 0;
	    }
	    break;
	case 24:
	    if (imread24(buf) != 0) {
		error("Error read data!!!");
		return 0;
	    }
	    break;
	default:
	    error("Bitmap file is not valid bit color");
    	    if(buf != NULL)
        	free(buf);
	    return 0;
        break;
    }
    
    /* Free memory */
    if(buf != NULL)
        free(buf);

    if (Lhs == 2) index_map_to_sci(colormap, 3);

    /* Return variables */
    LhsVar(1) = 2;
    LhsVar(2) = 3;
    return 0;
}
