/******************************/
/* Write array to image       */
/* imwrite(array, "filename") */
/******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "stack-c.h"
#include "common.h"

int imwrite(char *fname) {
    static int in;
    static int xin, yin, filename, bits, m2, n2;
    static int minlhs=1, maxlhs=1, minrhs=2, maxrhs=4;

    byte *buf; 			// Buffer
    HyperMat H; 		// Hyper matrix
    RGBquad *colormap;		// Color map
    byte rv=2;			// Right variable
    
    dword temp,lim;

    /* Check number of inputs and outputs */
    CheckRhs(minrhs,maxrhs);
    CheckLhs(minlhs,maxlhs);

    /* Get input array, filename and bits */
    switch (VarType(1)) {
	case 1:
	    GetRhsVar(1, "i", &xin, &yin, &in);
	    if (Rhs == 2) {
		for (temp = 0; temp < xin*yin; temp++) {
		    if (istk(in)[temp] > 1) {
			bits = 8;
			break;
		    }
		    else bits = 1;
		}
	    }
	    else bits = 8;
	    bih.biWidth=yin;				// Image width
	    bih.biHeight=xin;				// Image height
	    break;
	case 17:
	    bits = 24;
	    get_rhs_tru_img(1,&H,cstk(filename));
	    bih.biWidth = IC_INT32(H.sc.D)[1];
	    bih.biHeight = IC_INT32(H.sc.D)[0];
	break;
    }
    
    /* Read color map */
    if (Rhs == 3) {
	if ((colormap = (RGBquad *) malloc(pow(2,bits) * sizeof(RGBquad))) == NULL) {
	    error("Can't allocate memory for colormap\n");
	    if(buf != NULL)
	        free(buf);
	    return 0;
	}
	bits = sci_index_map_to_colormap(rv, colormap);
	rv++;
    }
    GetRhsVar(rv++, "c", &m2, &n2, &filename);

    /* Assign variables for bitmap file header */    
    bfh.bfType		= BMAGIC;			// BMP filetype
    bfh.bfReserved1	= 0;				// Unused
    bfh.bfReserved2	= 0;				// Unused
    bfh.bfOffBits	= 54+4*pow(2,bits);		// Data offset
    
    /* Assign variables for bitmap info header */    
    bih.biSize=40;					// Bitmap info header size
    bih.biPlanes=1;					// Number of Planes
    bih.biBitCount=bits;				// Bits per Pixel
    bih.biCompression=0;				// Type of Compression
    bih.biXPelsPerMeter=2925;				// Horizontal resolution: pixels/meter
    bih.biYPelsPerMeter=2925;				// Vertical resolution: pixels/meter
    bih.biClrUsed=pow(2,bits);				// Number of actually used colors
    bih.biClrImportant=pow(2,bits);			// Number of important colors

    /* Main part of programm - writing bitmap data in buffer */
    switch (bih.biBitCount) {
        case 1:
	    /* File size */
	    bfh.bfSize=bfh.bfOffBits+bih.biSizeImage;
	    /* Image size */
	    if (bih.biWidth <= 32) bih.biSizeImage=4*bih.biWidth;
		else bih.biSizeImage=bih.biWidth/32*4*bih.biHeight + bih.biHeight/8*32 + bih.biHeight%8*4;

	    if ((buf = (byte *) malloc(bih.biSizeImage * sizeof(byte))) == NULL) {
		error("Can't allocate memory for buffer\n");
	        if(buf != NULL)
		    free(buf);
		return 0;
	    }
	    if (imwrite1(buf, istk(in)) != 0) {
		error("Can't write data to buffer");
		return 0;
	    }
        break;
        case 4:
	    /* File size */
	    bfh.bfSize=bfh.bfOffBits+bih.biSizeImage;
	    /* Image size */
	    bih.biSizeImage=(bih.biWidth/2+4*((bih.biWidth/2)%8 || 0)-(bih.biWidth/2)%8)*bih.biHeight;

	    if ((buf = (byte *) malloc(bih.biSizeImage * sizeof(byte))) == NULL) {
		error("Can't allocate memory for buffer\n");
	        if(buf != NULL)
		    free(buf);
		return 0;
	    }
	    if (imwrite4(buf, istk(in)) != 0) {
		error("Can't write data to buffer");
		return 0;
	    }
        break;
        case 8:
	    /* File size */
	    bfh.bfSize=bfh.bfOffBits+bih.biSizeImage;
	    /* Image size */
	    bih.biSizeImage=(bih.biWidth+4*(bih.biWidth%4 || 0)-bih.biWidth%4)*bih.biHeight;
	    
	    if ((buf = (byte *) malloc(bih.biSizeImage * sizeof(byte))) == NULL) {
		error("Can't allocate memory for buffer\n");
	        if(buf != NULL)
		    free(buf);
		return 0;
	    }
	    if (imwrite8(buf, istk(in)) != 0) {
		error("Can't write data to buffer");
		return 0;
	    }
      break;
	case 24:
	    /* Determine main variables */
	    bfh.bfOffBits = 54;
	    bih.biClrUsed = 0;
	    bih.biClrImportant = 0;
	    bih.biSizeImage=bih.biWidth*bih.biHeight*3 + bih.biHeight*(bih.biWidth%4);
	    
	    /* File size */
	    bfh.bfSize=bfh.bfOffBits+bih.biSizeImage;
	    
	    if ((buf = (byte *) malloc(bih.biSizeImage * sizeof(byte))) == NULL) {
		error("Can't allocate memory for buffer\n");
	        if(buf != NULL)
		    free(buf);
		return 0;
	    }
	    if (imwrite24(buf, &H) != 0) {
		error("Can't write data to buffer");
		return 0;
	    }
	    break;
        default:
            error("Is not valid bits");
            return 0;
        break;
    }

    /* Open output file */
    if ((out = fopen(cstk(filename), "wb")) == NULL){
        error("Error opening output file");
	return 0;
    }

    /* Writing bitmap file header */
    if (fwrite(&bfh, sizeof(BitmapFileHeader), 1, out) != 1) {
	error("Error writing bitmap file header");
	return 0;
    }
    
    /* Writing bitmap info header */
    if (fwrite(&bih, sizeof(BitmapInfoHeader), 1, out) != 1) {
	error("Error writing bitmap info header");
	return 0;
    }
    
    /* Moving to color table segment */
    if (fseek(out, 54, SEEK_SET)) {
	error("Error writing color table");
	return 0;
    }

    /* Writing color table */
    switch(bih.biBitCount) {
        case 1:
        case 8:
	    if (Rhs == 2) {
        	temp = 0x000000;
        	for (lim=0; lim < pow(2,bih.biBitCount)-1; lim++) {
            	    fwrite(&temp, sizeof(dword), 1, out);
            	    temp += 0x010101;
        	}
        	temp = 0xFFFFFF;
        	fwrite(&temp, sizeof(dword), 1, out);
	    }
	    if (Rhs == 3) {
		if (fwrite(colormap, bih.biClrUsed*sizeof(RGBquad), 1, out) != 1) {
		    error("Error writing color map");
    		    return 0;
		}    
	    }
            break;
        case 4:
	    if (fwrite(colormap, bih.biClrUsed*sizeof(RGBquad), 1, out) != 1) {
		error("Error writing color map");
    		return 0;
	    }
            break;
	default:
	break;
    }

    /* Moving to data segment */
    if (fseek(out, bfh.bfOffBits, SEEK_SET)) {
	error("Error seeking to data segment");
	return 0;
    }
    
    if (fwrite(buf, sizeof(byte), bih.biSizeImage, out) != bih.biSizeImage) {
	error("Error writing image data");
	return 0;
    }

    /* Close output file */
    fclose(out);

    /* Return variables */
    LhsVar(1) = 0;
    return 0;
}
